require_relative 'installer_generator/version'
require_relative 'installer_generator/nsis_template'
require_relative 'installer_generator/binary'
require_relative 'installer_generator/exception'
require_relative 'installer_generator/cli'

require 'uri'
require 'fileutils'
require 'tmpdir'
require 'securerandom'
require 'pathname'

module Netfira
  class InstallerGenerator
    include FileUtils

    def self.preflight!
      [Binary::Makensis, Binary::Signcode].each do |klass|
        raise "No binaries available from [ #{klass.bin_files.join ' | '} ]" unless klass.available?
      end
    end

    # Because the writers are created en masse:
    # noinspection RubyResolve
    attr_reader :icon, :source, :signing_cert, :signing_key

    attr_accessor :target, :shop_id, :shop_url, :sync_url, :use_ssl,
                  :installer_name, :installer_filename, :temp_dir, :auth_server,
                  :signing_description, :signing_url,
                  :output_device

    REQUIRED_VALUES = %i[source icon target shop_id shop_url sync_url]

    def initialize
      self.output_device = :null
      self.use_ssl = true
      self.installer_name = 'Netfira Installer'
      self.installer_filename = 'netfira_dlm.exe'
      self.temp_dir = 'netfireb'
      self.auth_server = 'auth.netfira.biz'
    end

    {   icon: 'Icon',
        source: 'Source executable',
        signing_cert: 'Signing certificate',
        signing_key: 'Signing private key'

    }.each do |method, description|
      define_method :"#{method}=" do |path|
        raise Exception, "#{description} not found at #{path}" unless File.exist? path
        instance_variable_set :"@#{method}", path
      end
    end

    def generate!

      # Ensure we have all our binary friends
      self.class.preflight!

      # Ensure we have no missing required values
      require_values! REQUIRED_VALUES

      # Make a temporary build dir
      build_dir = Pathname(Dir.tmpdir) + SecureRandom.uuid
      mkdir build_dir

      begin

        # Copy the source exe and icon to the build dir
        cp icon, build_dir + 'setup.ico'
        cp source, build_dir + installer_filename

        # Write the NSIS script to the build dir
        File.write build_dir + 'nsis_script.nsi', render_nsis_template

        # Run makensis
        makensis build_dir or raise Exception, 'makensis failed'
        output_path = build_dir + 'nsis_generated.exe'

        # Run signcode
        if signing_key || signing_cert
          require_values! %i[signing_key signing_cert]
          input_path = output_path
          output_path = build_dir + 'signed.exe'
          signcode input_path, output_path or raise Exception, 'signcode failed'
        end

        # Write target file
        mkdir_p File.dirname(target)
        mv output_path, target

      ensure
        rm_rf build_dir
      end

    end

    private

    def require_values!(keys)
      missing = keys.select { |key| send(key).nil? }
      raise Exception, "Missing value(s) for #{missing.join ', '}" unless missing.empty?
    end

    def render_nsis_template
      # noinspection RubyStringKeysInHashInspection,SpellCheckingInspection
      values = {
          'INSTALLER_NAME' => installer_name,
          'TEMP_DIR' => temp_dir,
          'INSTALLER_FILENAME' => installer_filename,
          'SHOP_ID' => shop_id,
          'AUTHENTICATION_SERVER' => auth_server,
          'WEBSTORE_SYNC_URL' => sync_url,
          'WEBSTORE_FRONT_URL' => shop_url,
          'SHOP_ADDRESS' => URI.parse(sync_url).host,
          'NO_USE_SSL' => use_ssl ? '0' : '1'
      }
      NsisTemplate.sub values
    end

    def makensis(build_dir)
      b = Binary::Makensis.new
      b.template_file_name = 'nsis_script.nsi'
      b.build_dir = build_dir
      b.run output_device
    end

    def signcode(source, target)
      b = Binary::Signcode.new
      b.spc_path = signing_cert
      b.key_path = signing_key
      b.url = signing_url || shop_url
      b.description = signing_description || installer_name
      b.source_path = source
      b.target_path = target
      b.run output_device
    end

  end
end
